(* $Id: xstrp4_here.ml,v 1.3 2003/01/13 01:36:39 gerd Exp $
 * ----------------------------------------------------------------------
 *
 *)

open Xstrp4_here_types
open Pcaml


let interpolated_expr s =
  let rec parse_here_expr lexbuf =
    let tok = Xstrp4_here_lexer.token lexbuf in
      match tok with
	  Textend -> []
	| x       -> x :: parse_here_expr lexbuf
  in

  let rec normalize_literals =
    (* - Concat adjacent literals
     * - Remove empty literals
     *)
    function
	[] -> 
	  []
      | (Literal "") :: tl -> 
	  normalize_literals tl
      | (Literal s1) :: (Literal s2) :: tl -> 
	  normalize_literals((Literal(s1^s2))::tl)
      | hd :: tl ->
	  hd :: (normalize_literals tl)
  in

  let loc = (0, String.length s) in

  let buf = Lexing.from_string s in
  let toklist = normalize_literals (parse_here_expr buf) in

  let toklist_ast =
    List.map
      (function
	   Literal s -> <:expr< $str:s$ >>
	 | Variable (sl,fmt,pos1,pos2) -> 
	     let rec translate_id sl =
	       let loc = (pos1,pos2) in
	       match sl with
		   s :: ((s' :: _) as sl') -> 
		     let moduleid_ast = <:expr< $uid:s$ >> in
		     let valueid_ast = translate_id sl' in
		     <:expr< $moduleid_ast$ . $valueid_ast$ >>
		 | [s] ->
		     <:expr< $lid:s$ >>
		 | _ ->
		     failwith "Xstrp4.here_expr"
	     in
	     let node =  match fmt with
		 "%s" -> translate_id sl
	       | ("%d"|"%i") -> 
		   let id = translate_id sl in
		   <:expr< $lid:"string_of_int"$ $id$ >>
	       | _ ->
		   let id = translate_id sl in
		   <:expr< ( ( $uid:"Printf"$ . $lid:"sprintf"$ ) 
			     $str:fmt$
			   ) 
		           $id$ >>
	     in
	     node
	 | Textend -> failwith "Xstrp4.here_expr")
      toklist
  in

  let rec mk_list_ast l =
    match l with 
	[] -> <:expr< [] >>
      | x :: l' ->
	  let ast_l' = mk_list_ast l' in
	  <:expr< [ $x$ :: $ast_l'$ ] >>
  in

  let string_mod_ast =    <:expr< $uid:"String"$ >> in
  let concat_val_ast =    <:expr< $lid:"concat"$ >> in
  let string_concat_ast = <:expr< $string_mod_ast$ . $concat_val_ast$ >> in
  let concat_ast =        <:expr< $string_concat_ast$ $str:""$ >> in
  let list_ast =          mk_list_ast toklist_ast in
  let result_ast =        <:expr< $concat_ast$ $list_ast$ >> in

  match toklist with
      [] ->
	<:expr< $str:""$ >>
    | [Literal s] ->
	List.hd toklist_ast   (* = <:expr< $str:s$ >> *)
    | _ ->
	(* General case: *)
	result_ast
;;


let here_expr s =
  let result_ast = interpolated_expr s in
  let loc = (0, String.length s) in
  <:expr< $anti:result_ast$ >>
;;


let interpolated_file filename =
  let pathname =
    if Filename.is_implicit filename then
      Filename.concat (Filename.dirname !input_file) filename 
    else
      filename
  in
  let f = open_in pathname in
  let n = in_channel_length f in
  let s = String.create n in
  really_input f s 0 n;
  close_in f;
  here_expr s
;;


let included_file filename =
  let pathname =
    if Filename.is_implicit filename then
      Filename.concat (Filename.dirname !input_file) filename 
    else
      filename
  in
  let f = open_in pathname in
  let n = in_channel_length f in
  let s = String.create n in
  really_input f s 0 n;
  close_in f;
  let loc = (0, String.length s) in
  <:expr< $str:s$ >>
;;


let here_pat s =
  failwith "<<:here< >> documents not allowed in patterns"
;;


EXTEND
  expr: AFTER "simple"
    [[ "interpolate"; "file"; s = STRING -> interpolated_file s
     | "interpolate"; s = STRING -> interpolated_expr s 
     | "include"; "file"; s = STRING -> included_file s
     ]];
END
;;


Quotation.add
   "here"
   (Quotation.ExAst(here_expr, here_pat))
;;


(* ======================================================================
 * History:
 * 
 * $Log: xstrp4_here.ml,v $
 * Revision 1.3  2003/01/13 01:36:39  gerd
 * 	Improvement: Adajcent literals are merged, and emtpy
 * literal are removed (suggested by Mike Potanin)
 *
 * Revision 1.2  2001/04/21 16:53:15  gerd
 * 	Added "include file".
 *
 * Revision 1.1  1999/07/18 16:42:38  gerd
 * 	Initial revision.
 *
 * 
 *)
