(* $Id: xstrp4_here_lexer.mll,v 1.4 2002/12/28 18:49:51 gerd Exp $
 * ----------------------------------------------------------------------
 *
 *)

{
  open Xstrp4_here_types

  type val_id = LC of string | UC of string | End_of_id

  let rec parse_val_id f buf =
    let id = f buf in
      match id with
	  UC s -> s :: parse_val_id f buf 
	| LC s -> [s]
	| End_of_id -> []
}

let ucletter = [ 'A' - 'Z' ]
let lcletter = [ 'a' - 'z' '_' ] 
let acletter = ucletter | lcletter

let format = '%'
             [ '0' '-' ' ' ]*    (* no more modifiers are supported by Ocaml *)
             ['0'-'9']* 
             ( '.' ['0'-'9']* )?
	     ( ( ['L' 'l' 'n'] [ 'd' 'i' 'u' 'x' 'X' 'o' ])
	       | [ 'd' 'i' 'u' 'x' 'X' 's' 'c' 'f' 'e' 'E' 'g' 'G' 'b' 'a' 't' ]
	     )

rule token = parse
    '$' ( ucletter acletter* '.' )* lcletter acletter*
      { let s = Lexing.lexeme lexbuf in
	let buf = Lexing.from_string (String.sub s 1 (String.length s - 1)) in
	  Variable (parse_val_id value_identifier buf, 
		    "%s",
		   Lexing.lexeme_start lexbuf+1,
		   Lexing.lexeme_end lexbuf)
      }
  | '$' '{' ( ucletter acletter* '.' )* lcletter acletter*
            ( ',' format )?  
        '}'
      { let s = Lexing.lexeme lexbuf in
	let k_close = String.index s '}' in
	let k_percent = try String.index s '%' with Not_found -> (-1) in
	let buf = Lexing.from_string (String.sub s 2 (k_close - 1)) in
	let fmt = 
	  if k_percent >= 0 then
	    String.sub s k_percent (String.length s - k_percent - 1)
	  else
	    "%s"
	in
	let start = Lexing.lexeme_start lexbuf in
	Variable (parse_val_id value_identifier buf, 
		  fmt,
		  start+2,
		  start+(if k_percent >= 0 then k_percent-1 else k_close))
      }
  | '$'
      { failwith "Bad $ expander" }
  | '\\' '\n' 
      { Literal "" }
  | '\\' '$'
      { Literal "$" }
  | '\\' [ '0'-'9' ] [ '0'-'9' ] [ '0'-'9' ]
      {  Literal (let s = Lexing.lexeme lexbuf in
		  let n = int_of_string(String.sub s 1 3) in
                  Printf.sprintf "%c" (Char.chr n)
		 ) 
      }
(*
  | '\\' 'o' [ '0'-'7' ] [ '0'-'7' ] [ '0'-'7' ]
      {  Literal (let s = Lexing.lexeme lexbuf in
		  let n = int_of_string("0" ^ String.sub s 1 4) in
                  Printf.sprintf "%c" (Char.chr n)
		 ) 
      }
*)
  | '\\' 'x' [ '0'-'9' 'a'-'f' 'A'-'F' ] [ '0'-'9' 'a'-'f' 'A'-'F' ]
      {  Literal (let s = Lexing.lexeme lexbuf in
		  let n = int_of_string("0" ^ String.sub s 1 3) in
                  Printf.sprintf "%c" (Char.chr n)
		 ) 
      }

  | '\\' _
      { Literal (Lexing.lexeme lexbuf) }
  | [^ '$' '\\']+
      { Literal (Lexing.lexeme lexbuf) }
  | eof 
      { Textend }
  | _ 
      { Literal (Lexing.lexeme lexbuf) }

and value_identifier = parse
    ucletter acletter* '.'
      { let s = Lexing.lexeme lexbuf in
	  UC (String.sub s 0 (String.length s - 1))
      }
  | lcletter acletter*
      { LC(Lexing.lexeme lexbuf) }
  | eof
      { End_of_id }

(* ======================================================================
 * History:
 * 
 * $Log: xstrp4_here_lexer.mll,v $
 * Revision 1.4  2002/12/28 18:49:51  gerd
 * 	Removed backslash+octal
 *
 * Revision 1.3  2002/12/28 18:41:33  gerd
 * 	New backslash sequences
 *
 * Revision 1.2  2002/08/15 23:33:26  gerd
 * 	Added qualifiers for int32, int64, and nativeint.
 *
 * Revision 1.1  1999/07/18 16:42:39  gerd
 * 	Initial revision.
 *
 * 
 *)
